"""
Functions needed for backward compatibility.
"""


# Python 3.6+
try:
    from tld import get_fld, get_tld
    TLD_EXTRACTION, tldextract = None, None
# Python 3.5
except ImportError:
    import tldextract
    # extract callable that falls back to the included TLD snapshot, no live HTTP fetching
    TLD_EXTRACTION = tldextract.TLDExtract(suffix_list_urls=None)
    get_fld, get_tld = None, None



"""
Core functions needed to make the module work.
"""

## This file is available from https://github.com/adbar/courlan
## under GNU GPL v3 license

#import locale
import logging
import re
import sys

#from functools import cmp_to_key
from random import sample


from .clean import normalize_url, scrub_url
from .compatibility import TLD_EXTRACTION
from .filters import basic_filter, extension_filter, lang_filter, \
                     PATH_FILTER, spam_filter, type_filter, validate_url
from .network import redirection_test
from .settings import BLACKLIST
from .urlutils import extract_domain, fix_relative_urls, is_external


LOGGER = logging.getLogger(__name__)

FIND_LINKS_REGEX = re.compile(r'<a [^<>]+?>', re.I)
HREFLANG_REGEX = re.compile(r'hreflang=["\']([a-z-]+)', re.I)
LINK_REGEX = re.compile(r'href=["\']([^ ]+?)["\']', re.I)


def check_url(url, strict=False, with_redirects=False, language=None, with_nav=False):
    """ Check links for appropriateness and sanity
    Args:
        url: url to check
        strict: set to True for stricter filtering
        with_redirects: set to True for redirection test (per HTTP HEAD request)
        language: set target language (ISO 639-1 codes)
        with_nav: set to True to include navigation pages instead of discarding them

    Returns:
        A tuple consisting of canonical URL and extracted domain

    Raises:
        ValueError, handled in exception.
    """

    # first sanity check
    # use standard parsing library, validate and strip fragments, then normalize
    try:
        # length test
        if basic_filter(url) is False:
            raise ValueError

        # clean
        url = scrub_url(url)

        # get potential redirect
        if with_redirects is True:
            url = redirection_test(url)
            if url is None:
                raise ValueError

        # spam
        if strict is True and spam_filter(url) is False:
            raise ValueError
        # structural elements
        if type_filter(url, strict=strict, with_nav=with_nav) is False:
            raise ValueError

        # split and validate
        validation_test, parsed_url = validate_url(url)
        if validation_test is False:
            raise ValueError

        # content filter based on extensions
        if extension_filter(parsed_url.path) is False:
            raise ValueError

        # strict content filtering
        if strict is True and PATH_FILTER.match(parsed_url.path):
            raise ValueError

        # internationalization in URL
        if lang_filter(parsed_url.path, language) is False:
            raise ValueError

        # normalize
        url = normalize_url(parsed_url, strict, language)

    # handle exceptions
    except (AttributeError, ValueError, UnicodeError):
        # LOGGER.debug('discarded URL: %s', url)
        return None

    # domain info
    domain = extract_domain(url, blacklist=BLACKLIST)
    if domain is None:
        return None

    return url, domain


def sample_urls(urllist, samplesize, exclude_min=None, exclude_max=None, strict=False, verbose=False):
    '''Sample a list of URLs by domain name, optionally using constraints on their number'''
    lastseen, urlbuffer = None, set()
    if verbose is True:
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    else:
        logging.basicConfig(stream=sys.stdout, level=logging.ERROR)
    for url in sorted(urllist): # key=cmp_to_key(locale.strcoll)
        # first basic filter
        checked = check_url(url, strict=strict, with_redirects=False)
        if checked is None:
            continue
        url, domain = checked[0], checked[1]
        # continue collection
        if domain == lastseen:
            urlbuffer.add(url)
        # sample, drop, fresh start
        else:
            # threshold for too small websites
            if exclude_min is None or len(urlbuffer) >= exclude_min:
                # write all the buffer
                if len(urlbuffer) <= samplesize:
                    yield from sorted(urlbuffer)
                    LOGGER.info('%s\t\turls: %s', lastseen, len(urlbuffer))
                # or sample URLs
                else:
                    # threshold for too large websites
                    if exclude_max is None or len(urlbuffer) <= exclude_max:
                        yield from sorted(sample(urlbuffer, samplesize))
                        LOGGER.info('%s\t\turls: %s\tprop.: %s', lastseen, len(urlbuffer), samplesize/len(urlbuffer))
                    else:
                        LOGGER.info('discarded (exclude size): %s\t\turls: %s', lastseen, len(urlbuffer))
            else:
                LOGGER.info('discarded (exclude size): %s\t\turls: %s', lastseen, len(urlbuffer))
            urlbuffer = set()
            urlbuffer.add(url)
        lastseen = domain


def extract_links(pagecontent, base_url, external_bool, language=None,
                  strict=True, with_nav=False, redirects=False, reference=None):
    """ Filter links in a HTML document using a series of heuristics
    Args:
        pagecontent: whole page in binary format
        base_url: beginning of the URL, without path, fragment and query
        external: set to True for external links only, False for
                  internal links only
        language: set target language (ISO 639-1 codes)
        strict: set to True for stricter filtering
        with_nav: set to True to include navigation pages instead of discarding them
        with_redirects: set to True for redirection test (per HTTP HEAD request)
        reference: provide a host reference for external/internal evaluation

    Returns:
        A set containing filtered HTTP links checked for sanity and consistency.

    Raises:
        Nothing.
    """
    candidates, validlinks = set(), set()
    if pagecontent is None or pagecontent == '':
        return validlinks
    # define host reference
    if reference is None:
        if TLD_EXTRACTION is not None:
            reference = TLD_EXTRACTION(base_url)
        else:
            reference = base_url
    # extract links
    for link in FIND_LINKS_REGEX.findall(pagecontent):
        # https://en.wikipedia.org/wiki/Hreflang
        if language is not None and 'hreflang' in link:
            langmatch = HREFLANG_REGEX.search(link)
            if langmatch and (
                langmatch.group(1).startswith(language) or
                langmatch.group(1) == 'x-default'
                ):
                mymatch = LINK_REGEX.search(link)
                if mymatch:
                    candidates.add(mymatch.group(1))
        # default
        else:
            mymatch = LINK_REGEX.search(link)
            if mymatch:
                candidates.add(mymatch.group(1))
    # filter candidates
    for link in candidates:
        # repair using base
        if not link.startswith('http'):
            link = fix_relative_urls(base_url, link)
        # check
        checked = check_url(link, strict=strict, with_nav=with_nav,
                            with_redirects=redirects, language=language)
        if checked is None:
            continue
        # additional cleaning step?
        newlink = re.sub(r'/\&$', '', checked[0])
        # external/internal links
        if external_bool == is_external(newlink, reference):
            validlinks.add(newlink)
    # return
    LOGGER.info('%s links found – %s valid links', len(candidates), len(validlinks))
    return validlinks



"""
coURLan: Clean, filter, normalize, and sample URLs
"""


# meta
__title__ = 'courlan'
__author__ = 'Adrien Barbaresi'
__license__ = 'GNU GPL v3+'
__copyright__ = 'Copyright 2020-2021, Adrien Barbaresi'
__version__ = '0.4.2'


# imports
from .clean import clean_url, normalize_url, scrub_url
from .core import check_url, extract_links, sample_urls
from .filters import is_navigation_page, is_not_crawlable, lang_filter, validate_url
from .urlutils import extract_domain, fix_relative_urls, get_base_url, get_host_and_path, get_hostinfo, is_external



"""
Functions related to URL manipulation and extraction of URL parts.
"""

import re

from urllib.parse import urlparse


from .compatibility import TLD_EXTRACTION, get_fld, get_tld, tldextract


def extract_domain(url, blacklist=None):
    '''Extract domain name information using top-level domain info'''
    if blacklist is None:
        blacklist = set()
    # legacy tldextract code
    if TLD_EXTRACTION is not None:
        tldinfo = TLD_EXTRACTION(url)
        # domain TLD blacklist
        if tldinfo.domain in blacklist:
            return None
        # return domain
        returnval = '.'.join(part for part in tldinfo if part)
    # new code
    else:
        tldinfo = get_tld(url, as_object=True, fail_silently=True)
        # invalid input OR domain TLD blacklist
        if tldinfo is None or tldinfo.domain in blacklist:
            return None
        # return domain
        returnval = tldinfo.fld
    # this step seems necessary to standardize output
    return re.sub(r'^www[0-9]*\.', '', returnval)


def get_base_url(url):
    'Strip URL of some of its parts to get base URL.'
    parsed_url = urlparse(url)
    return parsed_url._replace(path='', params='', query='', fragment='').geturl()


def get_host_and_path(url):
    """Decompose URL in two parts: protocol + host/domain and path."""
    parsed_url = urlparse(url)
    host = parsed_url._replace(path='', params='', query='', fragment='')
    path = parsed_url._replace(scheme='', netloc='')
    return host.geturl(), path.geturl()


def get_hostinfo(url):
    """Extract domain and host info (protocol + host/domain) from a URL."""
    domainname = extract_domain(url)
    parsed_url = urlparse(url)
    host = parsed_url._replace(path='', params='', query='', fragment='')
    return domainname, host.geturl()


def fix_relative_urls(baseurl, url):
    'Prepend protocol and host information to relative links.'
    if url.startswith('//'):
        if baseurl.startswith('https'):
            urlfix = 'https:' + url
        else:
            urlfix = 'http:' + url
    elif url.startswith('/'):
        urlfix = baseurl + url
    # imperfect path handling
    elif url.startswith('.'):
        urlfix = baseurl + '/' + re.sub(r'(.+/)+', '', url)
    # don't try to correct these URLs
    elif url.startswith('{'):
        urlfix = url
    # catchall
    elif not url.startswith('http'):
        urlfix = baseurl + '/' + url
    else:
        urlfix = url
    return urlfix


def is_external(url, reference, ignore_suffix=True):
    '''Determine if a link leads to another host, takes a reference URL or
       tld/tldextract object as input, returns a boolean'''
    # legacy tldextract code
    if TLD_EXTRACTION is not None:
        # reference
        if not isinstance(reference, tldextract.tldextract.ExtractResult):
            reference = TLD_EXTRACTION(reference)
        tldinfo = TLD_EXTRACTION(url)
        if ignore_suffix is True:
            ref_domain, domain = reference.domain, tldinfo.domain
        else:  # '.'.join(ext[-2:]).strip('.')
            ref_domain, domain = reference.registered_domain, tldinfo.registered_domain
    # new tld code
    else:
        if ignore_suffix is True:
            try:
                ref_domain, domain = get_tld(reference, as_object=True, fail_silently=True).domain, \
                                     get_tld(url, as_object=True, fail_silently=True).domain
            # invalid input
            except AttributeError:
                return True
        else:
            ref_domain, domain = get_fld(reference, fail_silently=True), get_fld(url, fail_silently=True)
    # comparison
    if domain != ref_domain:
        return True
    return False



#!/usr/bin/python3
# -*- coding: utf-8 -*-

### Code made available on https://github.com/adbar/url-tools under GPL license.
### Adrien Barbaresi, 2015.

# Purpose: Revert URLs to their usual form


from __future__ import print_function

import argparse
import re
import time


parser = argparse.ArgumentParser()
parser.add_argument('-i', '--inputfile', dest='inputfile', help='name of the input file', required=True)
parser.add_argument('-o', '--outputfile', dest='outputfile', help='name of the output file', required=True)
args = parser.parse_args()

start_time = time.time()
lastseen = ''


with open(args.outputfile, 'w') as outputfh:
    with open(args.inputfile, 'r') as inputfh:
        for line in inputfh:
            # strip
            line = line.rstrip()
            # deduplicate
            if line is not lastseen:
                lastseen = line
                # reorder
                if re.search(r':http$', line):
                    line = re.sub(r':http$', '', line)

                    # normalize URL problem
                    line = re.sub(r'\./', '/', line, 1)

                    # not elegant, but will do the trick
                    match = re.match(r'(.+?)/', line)
                    if match:
                        core = match.group(1)
                    else:
                        print ('ERROR: Curious one (1):', lastseen)
                        continue
                    if not re.search(r'\.', core):
                        print ('ERROR: Curious one (2):', lastseen)
                        continue
                    elements = re.findall(r'.+?\.', core)
                    match = re.search(r'\.([^\.]+)$', core)
                    if match:
                        lastone = match.group(1)
                    else:
                        print ('ERROR: Curious one (3):', lastseen)
                        continue

                    # print the result of the substitution
                    core = ''
                    for item in reversed(elements):
                        core += item
                    core = re.sub(r'\.$', '', core)
                    line = re.sub(r'^.+?/', '/', line)
                    line = 'http://' + lastone + '.' + core + line
                    outputfh.write(line + '\n')


exec_time = time.time() - start_time
print ('# exec time:\t\t{0:.2f}' . format(exec_time))



"""
General settings for package execution.
"""

## This file is available from https://github.com/adbar/courlan
## under GNU GPL v3 license


# http://www.alexa.com/topsites/countries/DE
BLACKLIST = ('akamai', 'amzn', 'amazon', 'baidu', 'bit', 'bongacams', 'cloudfront', 'delicious', 'digg', 'ebay', 'ebay-kleinanzeigen', 'facebook', 'feedburner', 'flickr', 'gettyimages', 'gmx', 'google', 'gravatar', 'http', 'imgur', 'immobilienscout24', 'instagr', 'instagram', 'last', 'linkedin', 'live', 'livejasmin', 'localhost', 'mail', 'netflix', 'ok', 'otto', 'paypal', 'pinterest', 'pornhub', 'postbank', 'qq', 'reddit', 'sina', 'sohu', 'soundcloud', 'taobao', 'telegram', 'tmall', 'twitch', 'twitter', 'twitpic', 'txxx', 'vk', 'vkontakte', 'vimeo', 'web', 'weibo', 'whatsapp', 'xhamster', 'xvideos', 'yahoo', 'yandex', 'youtube', 'youtu')
# ...

ALLOWED_PARAMS = ('aid', 'article_id', 'artnr', 'id', 'itemid', 'objectid', 'p', 'page', 'pagenum', 'page_id', 'pid', 'post', 'postid', 'product_id')
CONTROL_PARAMS = ('lang', 'language')
TARGET_LANG_DE = ('de', 'deutsch', 'ger', 'german')
TARGET_LANG_EN = ('en', 'english', 'eng') # 'en_US', ''
# accepted_lang = ('en')



#!/usr/bin/python3
# -*- coding: utf-8 -*-

### Code made available on https://github.com/adbar/url-tools under GPL license.
### Adrien Barbaresi, 2015.

# Purpose: Extract possible WordPress links
# http://codex.wordpress.org/Using_Permalinks

## TODO:
# fasttrack: wordpress.com
# tests: nebojsaozimic potkozarje


from __future__ import print_function
from urllib.parse import urlparse
import re
import argparse
import time

parser = argparse.ArgumentParser()
parser.add_argument('-i', '--inputfile', dest='inputfile', help='name of the input file', required=True)
parser.add_argument('-o', '--outputfile', dest='outputfile', help='name of the output file', required=True)
parser.add_argument('-l', '--lax', dest='lax', action="store_true", default=False, help='use lax patterns')
args = parser.parse_args()

start_time = time.time()
known_hosts = set()
i = 0
j = 0
k = 0
l = 0


# test if part of the URL is interesting
def find_target(url):
    # uparse = urlparse(url)

    # wordpress.com
    if re.match(r'https?://.+?\.wordpress\.[a-z]{2,3}', url): # [a-z]+?
       # files.wordpress.com hack
       url = re.sub(r'\.files\.wordpress\.', '.wordpress.', url)
       # wphost = re.match (r'(htt.+?\.wordpress\.[a-z]{2,3}/)', url)
       wphost = re.match (r'(htt.+?\.wordpress\.[a-z]{2,3})/?', url)
       if wphost:
           return wphost.group(1).rstrip('/') + '/'

    # K.O. victory
    ko_string = re.match(r'(.+?)(/wp/|/wordpress/|/wp-content/)', url)
    if ko_string:
        return ko_string.group(1).rstrip('/') + '/'

    # /tag/ /category/
    tagcat = re.match(r'(.+?)(/tag/|/category/|\?cat=)', url)
    if tagcat:
        return tagcat.group(1).rstrip('/') + '/'

    # query parameters
    if re.search(r'/\?p=|\?page_id=|\?paged=/', url):
        mquery = re.match(r'(https?://.+?/)(blog/|weblog/)?(\?p=|\?page_id=|\?paged=)', url)
        if mquery:
            if mquery.group(2) and mquery.group(3):
                return mquery.group(1) + mquery.group(2)
            else:
                return mquery.group(1).rstrip('/') + '/'

    # URL types
    if re.search(r'/20[0-9]{2}/[0-9]{2}/|/archives/', url):
        url_types = re.search(r'(https?://.+?/)(blog/|weblog/)?(20[0-9]{2}/[0-9]{2}/|/archives/)', url)
        if url_types:
            # print (url)
            if url_types.group(2) and url_types.group(3):
                return url_types.group(1) + url_types.group(2)
            else:
                return url_types.group(1).rstrip('/') + '/'

    # lax
    if args.lax is True:
    # path correction
    # mpath = re.match(r'(/blog/|/weblog/)', url) #uparse.path
    # if mpath:
    #    path = mpath.group(1)
    #else:
    #    path = '' 
        if re.search(r'/[a-z]+-[a-z]+-[a-z]+|/20[0-9]{2}/', url):
            url_lax = re.search(r'(https?://.+?/)(blog/|weblog/)?(/[a-z]+-[a-z]+-[a-z]+|/20[0-9]{2}/)', url)
            if url_lax:
                if url_lax.group(2) and url_lax.group(3):
                    return url_lax.group(1) + url_lax.group(2)
                else:
                    return url_lax.group(1).rstrip('/') + '/'

    return None

with open(args.inputfile, 'r', encoding='utf-8') as inputfh:
    with open(args.outputfile, 'w', encoding='utf-8') as outputfh:
        # avoid errors (todo: solve)
        try:
            for url in inputfh:
                i += 1
                target = None
                # sanitize
                url = url.lower().rstrip('\n')

                # filters
                if re.match('http', url) and len(url) > 11:
                    # akamai/fbcdn, etc.
                    if not re.search(r'\.blogspot\.|\.google\.|\.tumblr\.|\.typepad\.com|\.wp\.com|\.archive\.|akamai|fbcdn|baidu\.com|\.gravatar\.', url):
                        # test if part of the URL is interesting
                        target = find_target(url)

                # limit path depth and filter out queries
                if target and not re.search(r'=|\.php', target) and len(re.findall(r'/', target)) <= 4:
                    # make sure the host name is fresh
                    normalize = re.search(r'https?://(www\.)?(.+?)/', target)
                    if normalize and normalize.group(2) not in known_hosts:
                    # if True:
                        known_hosts.add(normalize.group(2))
                        outputfh.write(target + '\n')
                        j += 1
        except UnicodeDecodeError:
            print ('Unicode error line', i + 1)


print ('# lines seen:\t\t', i)
print ('# results:\t\t', j)
exec_time = time.time() - start_time
print ('# exec time:\t\t{0:.2f}' . format(exec_time))
print ('# secs per success:\t{0:.2f}' . format(exec_time / j))
# print (k, l)



"""
Bundles functions needed to target text content and validate the input.
"""


## This file is available from https://github.com/adbar/courlan
## under GNU GPL v3 license

import re

from urllib.parse import urlparse


# content filters
WORDPRESS_CONTENT_FILTER = re.compile(r'/(?:page|seite|user|search|gallery|gall?erie|labels|archives|uploads|modules|attachment)/', re.IGNORECASE)
WORDPRESS_CONTENT_FILTER_2 = re.compile(r'/(?:tags?|schlagwort|category|cat|kategorie|kat|auth?or)/[^/]+$', re.IGNORECASE)
PARAM_FILTER = re.compile(r'\.(atom|json|css|xml|js|jpg|jpeg|png|gif|tiff|pdf|ogg|mp3|m4a|aac|avi|mp4|mov|webm|flv|ico|pls|zip|tar|gz|iso|swf)\b', re.IGNORECASE)  # , re.IGNORECASE (?=[&?])
PATH_FILTER = re.compile(r'.{0,5}/(impressum|index)(\.[a-z]{3,4})?/?$', re.IGNORECASE)
ADULT_FILTER = re.compile(r'\b(?:adult|amateur|arsch|cams?|cash|fick|gangbang|incest|porn|sexyeroti[ck]|sexcam|swinger|xxx|bild\-?kontakte)\b', re.IGNORECASE) # live|sex|ass|orgasm|cams|

# language filter
URL_LANG_FILTER = re.compile(r'/([a-z]{2,3})([_-][A-Za-z]{2,3})?/', re.IGNORECASE)

# navigation/crawls
NAVIGATION_FILTER = re.compile(r'/(archives|auth?or|cat|category|kat|kategorie|page|schlagwort|seite|tags?|topics?|user)/', re.IGNORECASE) # ?p=[0-9]+$
NOTCRAWLABLE = re.compile(r'/(login|impressum|imprint)/?$|/login\?|/(javascript:|mailto:|tel\.?:|whatsapp:)', re.IGNORECASE)
# |/(www\.)?(facebook\.com|google\.com|instagram\.com|twitter\.com)/

# document types
WHITELISTED_EXTENSIONS = ('.amp', '.asp', '.aspx', '.cfm', '.cgi', '.htm', 'html', '.shtml', '.jsp', '.php', '.pl', '.txt')


def basic_filter(url):
    '''Filter URLs based on basic formal characteristics'''
    if not url.startswith('http') or len(url) >= 500 or len(url) < 10:
        return False
    return True


def extension_filter(urlpath):
    '''Filter based on file extension'''
    if re.search(r'\.[a-z]{2,5}$', urlpath) and not urlpath.endswith(WHITELISTED_EXTENSIONS):
        return False
    return True


def lang_filter(url, language):
    '''Heuristic targeting internationalization'''
    if language is not None:
        match = URL_LANG_FILTER.search(url)
        if match:
            if language == 'de' and match.group(1).lower() not in ('de', 'deu'):
                return False
            elif language == 'en' and match.group(1).lower() not in ('en', 'eng'):
                return False
            elif language != match.group(1).lower()[:2]:
                return False
    return True


def spam_filter(url):
    '''Try to filter out spam and adult websites'''
    # TODO: to improve!
    #for exp in (''):
    #    if exp in url:
    #        return False
    if ADULT_FILTER.search(url):
        return False
    # default
    return True


def type_filter(url, strict=False, with_nav=False):
    '''Make sure the target URL is from a suitable type (HTML page with primarily text)'''
    # directory
    #if url.endswith('/'):
    #    return False
    try:
        # feeds
        if url.endswith(('/feed', '/rss')):
            raise ValueError
        # embedded content
        if re.search(r'/oembed\b', url, re.IGNORECASE):
            raise ValueError
        # wordpress structure
        if WORDPRESS_CONTENT_FILTER.search(url) or WORDPRESS_CONTENT_FILTER_2.search(url):
            if with_nav is not True or not is_navigation_page(url):
                raise ValueError
        # hidden in parameters
        if strict is True and PARAM_FILTER.search(url):
            raise ValueError
        # not suitable
        if re.match(r'https?://banner\.|https?://add?s?\.', url, re.IGNORECASE):
            raise ValueError
        if re.search(r'\b(?:doubleclick|tradedoubler|livestream)\b|/(?:live|videos?)/', url, re.IGNORECASE):
            raise ValueError
        if strict is True and re.search(r'\b(?:live|videos?)\b', url, re.IGNORECASE):
            raise ValueError
    except ValueError:
        return False
    # default
    return True


def validate_url(url):
    '''Parse and validate the input'''
    try:
        parsed_url = urlparse(url)
    except ValueError:
        return False, None
    if bool(parsed_url.scheme) is False or parsed_url.scheme not in ('http', 'https'):
        return False, None
    if len(parsed_url.netloc) < 5 or \
       (parsed_url.netloc.startswith('www.') and len(parsed_url.netloc) < 8):
        return False, None
    # if validators.url(parsed_url.geturl(), public=True) is False:
    #    return False
    # default
    return True, parsed_url


def is_navigation_page(url):
    '''Determine if the URL is related to navigation and overview pages
       rather than content pages, e.g. /page/1 vs. article page.'''
    return bool(NAVIGATION_FILTER.search(url))


def is_not_crawlable(url):
    '''Run tests to check if the URL may lead to deep web or pages
       generally not usable in a crawling context.'''
    return bool(NOTCRAWLABLE.search(url))



"""
Unit tests for the courlan package.
"""

## This file is available from https://github.com/adbar/courlan
## under GNU GPL v3 license

import logging
import os
import sys

from unittest.mock import patch

import pytest

try:
    import tldextract
    TLD_EXTRACTION = tldextract.TLDExtract(suffix_list_urls=None)
except ImportError:
    TLD_EXTRACTION = None

from courlan import clean_url, normalize_url, scrub_url, check_url, is_external, sample_urls, validate_url, extract_links, extract_domain, fix_relative_urls, get_base_url, get_host_and_path, get_hostinfo, is_navigation_page, is_not_crawlable, lang_filter
from courlan.cli import parse_args
from courlan.filters import extension_filter, spam_filter, type_filter


logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)



def test_baseurls():
    assert get_base_url('https://example.org/') == 'https://example.org'
    assert get_base_url('https://example.org/test.html?q=test#frag') == 'https://example.org'
    assert get_base_url('example.org') == ''


def test_fix_relative():
    assert fix_relative_urls('https://example.org', 'page.html') == 'https://example.org/page.html'
    assert fix_relative_urls('http://example.org', '//example.org/page.html') == 'http://example.org/page.html'
    assert fix_relative_urls('https://example.org', './page.html') == 'https://example.org/page.html'
    assert fix_relative_urls('https://example.org', '/page.html') == 'https://example.org/page.html'
    # fixing partial URLs
    assert fix_relative_urls('https://example.org', 'https://example.org/test.html') == 'https://example.org/test.html'
    assert fix_relative_urls('https://example.org', '/test.html') == 'https://example.org/test.html'
    assert fix_relative_urls('https://example.org', '//example.org/test.html') == 'https://example.org/test.html'
    assert fix_relative_urls('http://example.org', '//example.org/test.html') == 'http://example.org/test.html'
    assert fix_relative_urls('https://example.org', 'test.html') == 'https://example.org/test.html'
    assert fix_relative_urls('https://example.org', '../../test.html') == 'https://example.org/test.html'


def test_scrub():
    # clean: scrub + normalize
    assert clean_url(5) is None
    assert clean_url('ø\xaa') == 'øª'
    # scrub
    assert scrub_url('  https://www.dwds.de') == 'https://www.dwds.de'
    assert scrub_url('<![CDATA[https://www.dwds.de]]>') == 'https://www.dwds.de'
    assert scrub_url('https://www.dwds.de/test?param=test&amp;other=test') == 'https://www.dwds.de/test?param=test&other=test'
    assert scrub_url('https://www.dwds.de/garbledhttps://www.dwds.de/') == 'https://www.dwds.de/garbled'
    assert scrub_url('https://g__https://www.dwds.de/') == 'https://www.dwds.de'
    # exception for archive URLs
    assert scrub_url('https://web.archive.org/web/20131021165347/https://www.imdb.com/') == 'https://web.archive.org/web/20131021165347/https://www.imdb.com'
    # social sharing
    assert scrub_url('https://twitter.com/share?&text=Le%20sabre%20de%20bambou%20%232&via=NouvellesJapon&url=https://nouvellesdujapon.com/le-sabre-de-bambou-2') == 'https://nouvellesdujapon.com/le-sabre-de-bambou-2'
    assert scrub_url('https://www.facebook.com/sharer.php?u=https://nouvellesdujapon.com/le-sabre-de-bambou-2') == 'https://nouvellesdujapon.com/le-sabre-de-bambou-2'
    # end of URL
    assert scrub_url('https://www.test.com/&') == 'https://www.test.com'
    # white space
    assert scrub_url('\x19https://www.test.com/\x06') == 'https://www.test.com'
    # markup
    assert scrub_url('https://www.test.com/</a>') == 'https://www.test.com'
    # garbled URLs e.g. due to quotes
    assert scrub_url('https://www.test.com/"' + '<p></p>'*100) == 'https://www.test.com'
    assert scrub_url('https://www.test.com/"' * 50) != 'https://www.test.com'


def test_extension_filter():
    assert extension_filter('http://www.example.org/test.js') is False
    assert extension_filter('http://goodbasic.com/GirlInfo.aspx?Pseudo=MilfJanett') is True
    assert extension_filter('https://www.familienrecht-allgaeu.de/de/vermoegensrecht.amp') is True
    assert extension_filter('http://www.example.org/test.shtml') is True


def test_spam_filter():
    assert spam_filter('http://www.example.org/cams/test.html') is False
    assert spam_filter('http://www.example.org/test.html') is True


def test_type_filter():
    assert type_filter('http://www.example.org/feed') is False
    # straight category
    assert type_filter('http://www.example.org/category/123') is False
    # post simply filed under a category
    assert type_filter('http://www.example.org/category/tropes/time-travel') is True
    assert type_filter('http://www.example.org/test.xml?param=test', strict=True) is False
    assert type_filter('http://www.example.org/test.asp') is True
    assert type_filter('http://ads.example.org/') is False
    # -video- vs. /video/
    assert type_filter('http://my-videos.com/') is True
    assert type_filter('http://my-videos.com/', strict=True) is False
    assert type_filter('http://example.com/video/1') is False
    assert type_filter('http://example.com/new-video-release') is True
    assert type_filter('http://example.com/new-video-release', strict=True) is False


def test_lang_filter():
    assert lang_filter('https://www.20min.ch/fr/story/des-millions-pour-produire-de-l-energie-renouvelable-467974085377', None) is True
    assert lang_filter('https://www.20min.ch/fr/story/des-millions-pour-produire-de-l-energie-renouvelable-467974085377', 'de') is False
    assert lang_filter('https://www.20min.ch/fr/story/des-millions-pour-produire-de-l-energie-renouvelable-467974085377', 'fr') is True
    assert lang_filter('https://www.20min.ch/fr/story/des-millions-pour-produire-de-l-energie-renouvelable-467974085377', 'en') is False
    assert lang_filter('https://www.20min.ch/fr/story/des-millions-pour-produire-de-l-energie-renouvelable-467974085377', 'es') is False
    assert lang_filter('https://www.sitemaps.org/en_GB/protocol.html', 'en') is True
    assert lang_filter('https://www.sitemaps.org/en_GB/protocol.html', 'de') is False


def test_navigation():
    assert is_navigation_page('https://test.org/') is False
    assert is_navigation_page('https://test.org/page/1') is True
    assert is_not_crawlable('https://test.org/login') is True
    assert is_not_crawlable('https://test.org/login/') is True
    assert is_not_crawlable('https://test.org/page') is False


def test_validate():
    assert validate_url('http://www.test[.org/test')[0] is False
    # assert validate_url('http://www.test.org:7ERT/test')[0] is False
    assert validate_url('ntp://www.test.org/test')[0] is False
    assert validate_url('ftps://www.test.org/test')[0] is False
    assert validate_url('http://t.g/test')[0] is False
    assert validate_url('http://test.org/test')[0] is True


def test_normalization():
    assert normalize_url('HTTPS://WWW.DWDS.DE/') == 'https://www.dwds.de/'
    assert normalize_url('http://test.net/foo.html#bar', strict=True) == 'http://test.net/foo.html'
    assert normalize_url('http://test.net/foo.html#bar', strict=False) == 'http://test.net/foo.html#bar'
    assert normalize_url('http://test.net/foo.html#:~:text=night-,vision', strict=True) == 'http://test.net/foo.html'
    assert normalize_url('http://www.example.org:80/test.html') == 'http://www.example.org/test.html'
    assert normalize_url('https://hanxiao.io//404.html') == 'https://hanxiao.io/404.html'


def test_qelems():
    assert normalize_url('http://test.net/foo.html?utm_source=twitter') == 'http://test.net/foo.html?utm_source=twitter'
    assert normalize_url('http://test.net/foo.html?utm_source=twitter', strict=True) == 'http://test.net/foo.html'
    assert normalize_url('http://test.net/foo.html?utm_source=twitter&post=abc&page=2') == 'http://test.net/foo.html?page=2&post=abc&utm_source=twitter'
    assert normalize_url('http://test.net/foo.html?utm_source=twitter&post=abc&page=2', strict=True) == 'http://test.net/foo.html?page=2&post=abc'
    assert normalize_url('http://test.net/foo.html?page=2&itemid=10&lang=en') == 'http://test.net/foo.html?itemid=10&lang=en&page=2'
    with pytest.raises(ValueError):
        assert normalize_url('http://test.net/foo.html?page=2&lang=en', language='de')
        assert normalize_url('http://www.evolanguage.de/index.php?page=deutschkurse_fuer_aerzte&amp;language=ES', language='de')


def test_urlcheck():
    assert check_url('AAA') is None
    assert check_url('1234') is None
    assert check_url('http://ab') is None
    assert check_url('ftps://example.org/') is None
    assert check_url('http://t.g/test') is None
    assert check_url('https://www.dwds.de/test?param=test&amp;other=test', strict=True) == ('https://www.dwds.de/test', 'dwds.de')
    assert check_url('http://example.com/index.html#term', strict=True) is None
    assert check_url('http://example.com/index.html#term', strict=False)[0] == 'http://example.com/index.html#term'
    assert check_url('http://example.com/test.js') is None
    assert check_url('http://twitter.com/') is None
    assert check_url('https://www.httpbin.org/status/200', with_redirects=True) == ('https://www.httpbin.org/status/200', 'httpbin.org')
    #assert check_url('https://www.httpbin.org/status/302', with_redirects=True) == ('https://www.httpbin.org/status/302', 'httpbin.org')
    assert check_url('https://www.httpbin.org/status/404', with_redirects=True) is None
    assert check_url('https://www.ht.or', with_redirects=True) is None
    if TLD_EXTRACTION is None:
        assert check_url('http://www.example') is None
        assert check_url('http://example.invalid/', False) is None
    # recheck type and spam filters
    assert check_url('http://example.org/code/oembed/') is None
    assert check_url('http://cams.com/', strict=False) == ('http://cams.com', 'cams.com')
    assert check_url('http://cams.com/', strict=True) is None
    assert check_url('https://denkiterm.wordpress.com/impressum/', strict=True) is None
    assert check_url('http://www.fischfutter-index.de/improvit-trocken-frostfutter-fur-fast-alle-fische/', strict=True) is not None
    # language and internationalization
    assert check_url('http://example.com/test.html?lang=en', language='de') is None
    assert check_url('http://example.com/test.html?lang=en', language=None) is not None
    assert check_url('http://example.com/test.html?lang=en', language='en') is not None
    assert check_url('http://example.com/de/test.html', language='de') is not None
    assert check_url('http://example.com/en/test.html', language='de') is None
    assert check_url('http://example.com/en/test.html', language=None) is not None
    assert check_url('http://example.com/en/test.html', language='en') is not None
    assert check_url('https://www.myswitzerland.com/de-ch/erlebnisse/veranstaltungen/wild-im-sternen/', language='de') is not None
    assert check_url('https://www.myswitzerland.com/en-id/accommodations/other-types-of-accommodations/on-the-farm/farm-experiences-search/', language='en') is not None
    assert check_url('https://www.myswitzerland.com/EN-ID/accommodations/other-types-of-accommodations/on-the-farm/farm-experiences-search/', language='en') is not None
    # impressum and index
    assert check_url('http://www.example.org/index', strict=True) is None
    assert check_url('http://www.example.org/index.html', strict=True) is None
    assert check_url('http://concordia-hagen.de/impressum.html', strict=True) is None
    assert check_url('http://concordia-hagen.de/de/impressum', strict=True) is None
    assert check_url('http://parkkralle.de/detail/index/sArticle/2704', strict=True) is not None
    assert check_url('https://www.katholisch-in-duisdorf.de/kontakt/links/index.html', strict=True) is not None
    assert check_url('{mylink}') is None


def test_urlutils():
    '''Test URL manipulation tools'''
    assert extract_domain('https://httpbin.org/') == 'httpbin.org'
    assert get_base_url('https://example.org/path') == 'https://example.org'
    assert get_host_and_path('https://example.org/path') == ('https://example.org', '/path')
    assert get_hostinfo('https://example.org/path') == ('example.org', 'https://example.org')
    assert get_hostinfo('https://httpbin.org/') == ('httpbin.org', 'https://httpbin.org')


def test_external():
    '''test domain comparison'''
    assert is_external('https://github.com/', 'https://www.microsoft.com/') is True
    assert is_external('https://microsoft.com/', 'https://www.microsoft.com/', ignore_suffix=True) is False
    assert is_external('https://microsoft.com/', 'https://www.microsoft.com/', ignore_suffix=False) is False
    assert is_external('https://google.com/', 'https://www.google.co.uk/', ignore_suffix=True) is False
    assert is_external('https://google.com/', 'https://www.google.co.uk/', ignore_suffix=False) is True
    # malformed URLs
    assert is_external('h1234', 'https://www.google.co.uk/', ignore_suffix=True) is True
    #if TLD_EXTRACTION is not None:
    #    # tldextract object
    #    tldinfo = TLD_EXTRACTION('http://127.0.0.1:8080/test/')
    #    assert is_external('https://127.0.0.1:80/', tldinfo) is False


def test_extraction():
    '''test link comparison in HTML'''
    assert len(extract_links(None, 'https://test.com/', False)) == 0
    assert len(extract_links('', 'https://test.com/', False)) == 0
    # language
    pagecontent = '<html><a href="https://test.com/example" hreflang="de-DE"/></html>'
    assert len(extract_links(pagecontent, 'https://test.com/', False)) == 1
    assert len(extract_links(pagecontent, 'https://test.com/', True)) == 0
    assert len(extract_links(pagecontent, 'https://test.com/', False, language='de')) == 1
    assert len(extract_links(pagecontent, 'https://test.com/', False, language='en')) == 0
    # x-default
    pagecontent = '<html><a href="https://test.com/example" hreflang="x-default"/></html>'
    assert len(extract_links(pagecontent, 'https://test.com/', False, language='de')) == 1
    assert len(extract_links(pagecontent, 'https://test.com/', False, language='en')) == 1
    # language + content
    pagecontent = '<html><a hreflang="de-DE" href="https://test.com/example"/><a href="https://test.com/example2"/><a href="https://test.com/example2 ADDITIONAL"/></html>'
    links = extract_links(pagecontent, 'https://test.com/', False)
    assert sorted(links) == ['https://test.com/example', 'https://test.com/example2']
    assert len(extract_links(pagecontent, 'https://test.com/', False, language='de')) == 2
    pagecontent = '<html><a hreflang="de-DE" href="https://test.com/example"/><a href="https://test.com/page/2"/></html>'
    assert len(extract_links(pagecontent, 'https://test.com/', False, with_nav=False)) == 1
    assert len(extract_links(pagecontent, 'https://test.com/', False, with_nav=True)) == 2
    pagecontent = "<html><head><title>Links</title></head><body><a href='/links/2/0'>0</a> <a href='/links/2/1'>1</a> </body></html>"
    links = extract_links(pagecontent, 'https://httpbin.org', False, with_nav=True)
    assert sorted(links) == ['https://httpbin.org/links/2/0', 'https://httpbin.org/links/2/1']
    # links undeveloped by CMS
    pagecontent = '<html><a href="{privacy}" target="_privacy">{privacy-link}</a></html>'
    assert len(extract_links(pagecontent, 'https://test.com/', False)) == 0
    assert len(extract_links(pagecontent, 'https://test.com/', True)) == 0


def test_cli():
    '''test the command-line interface'''
    testargs = ['', '-i', 'input.txt', '--outputfile', 'output.txt', '-v', '--language', 'en']
    with patch.object(sys, 'argv', testargs):
        args = parse_args(testargs)
    assert args.inputfile == 'input.txt'
    assert args.outputfile == 'output.txt'
    assert args.verbose is True
    assert args.language == 'en'
    assert os.system('courlan --help') == 0  # exit status


def test_sample():
    '''test URL sampling'''
    assert len(list(sample_urls(['http://test.org/test1', 'http://test.org/test2'], 0))) == 0
    # assert len(sample_urls(['http://test.org/test1', 'http://test.org/test2'], 1)) == 1
    mylist = ['http://t.o/t1', 'http://test.org/test1', 'http://test.org/test2', 'http://test2.org/test2']
    assert len(list(sample_urls(mylist, 1, verbose=True))) == 1
    assert len(list(sample_urls(mylist, 1, exclude_min=10, verbose=True))) == 0
    assert len(list(sample_urls(mylist, 1, exclude_max=1, verbose=True))) == 0


def test_examples():
    '''test README examples'''
    assert check_url('https://github.com/adbar/courlan') == ('https://github.com/adbar/courlan', 'github.com')
    assert check_url('https://httpbin.org/redirect-to?url=http%3A%2F%2Fexample.org', strict=True) == ('https://httpbin.org/redirect-to', 'httpbin.org')
    assert clean_url('HTTPS://WWW.DWDS.DE:80/') == 'https://www.dwds.de'
    assert validate_url('http://1234') == (False, None)
    assert validate_url('http://www.example.org/')[0] is True
    assert normalize_url('http://test.net/foo.html?utm_source=twitter&post=abc&page=2#fragment', strict=True) == 'http://test.net/foo.html?page=2&post=abc'



"""
Functions performing URL trimming and cleaning
"""

## This file is available from https://github.com/adbar/courlan
## under GNU GPL v3 license

import logging
import re

from collections import OrderedDict
from urllib.parse import parse_qs, urlencode, urlparse, ParseResult

from .filters import validate_url
from .settings import ALLOWED_PARAMS, CONTROL_PARAMS,\
                      TARGET_LANG_DE, TARGET_LANG_EN


PROTOCOLS = re.compile(r'https?://')
SELECTION = re.compile(r'(https?://[^">&? ]+?)(?:https?://)|(?:https?://[^/]+?/[^/]+?[&?]u(rl)?=)(https?://[^"> ]+)')
MIDDLE_URL = re.compile(r'https?://.+?(https?://.+?)(?:https?://|$)')

NETLOC_RE = re.compile(r'(?<=\w):(?:80|443)')
PATH1 = re.compile(r'/+')
PATH2 = re.compile(r'^(?:/\.\.(?![^/]))+')


def clean_url(url, language=None):
    '''Helper function: chained scrubbing and normalization'''
    try:
        return normalize_url(scrub_url(url), language)
    except (AttributeError, ValueError):
        return None


def scrub_url(url):
    '''Strip unnecessary parts and make sure only one URL is considered'''
    # trim
    # https://github.com/cocrawler/cocrawler/blob/main/cocrawler/urls.py
    # remove leading and trailing white space and unescaped control chars
    url = url.strip('\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
                    '\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f \r\n')
    # clean the input string
    url = url.replace('[ \t]+', '')
    # <![CDATA[http://www.urbanlife.de/item/260-bmw-i8-hybrid-revolution-unter-den-sportwagen.html]]>
    if url.startswith('<![CDATA['): # re.match(r'<!\[CDATA\[', url):
        url = url.replace('<![CDATA[', '') # url = re.sub(r'^<!\[CDATA\[', '', url)
        url = url.replace(']]>', '') # url = re.sub(r'\]\]>$', '', url)
    # markup rests
    url = re.sub(r'</?a>', '', url)
    # &amp;
    if '&amp;' in url:
        url = url.replace('&amp;', '&')
    #if '"' in link:
    #    link = link.split('"')[0]
    # double/faulty URLs
    protocols = PROTOCOLS.findall(url)
    if len(protocols) > 1 and not 'web.archive.org' in url:
        logging.debug('double url: %s %s', len(protocols), url)
        match = SELECTION.match(url)
        if match and validate_url(match.group(1))[0] is True:
            url = match.group(1)
            logging.debug('taking url: %s', url)
        else:
            match = MIDDLE_URL.match(url)
            if match and validate_url(match.group(1))[0] is True:
                url = match.group(1)
                logging.debug('taking url: %s', url)
    # too long and garbled URLs e.g. due to quotes URLs
    # https://github.com/cocrawler/cocrawler/blob/main/cocrawler/urls.py
    if len(url) > 500:  # arbitrary choice
        match = re.match(r'(.*?)[<>"\'\r\n ]', url)
        if match:
            url = match.group(1)
        if len(url) > 500:
            logging.debug('invalid-looking link %s of length %d',
                           url[:50] + '...', len(url))
    # trailing ampersand
    url = url.strip('&')
    # trailing slashes in URLs without path or in embedded URLs
    if url.count('/') == 3 or url.count('://') > 1:
        url = url.rstrip('/')
    # lower
    # url = url.lower()
    return url


def clean_query(parsed_url, strict=False, language=None):
    '''Strip unwanted query elements'''
    if len(parsed_url.query) > 0:
        qdict = parse_qs(parsed_url.query)
        newqdict = OrderedDict()
        for qelem in sorted(qdict.keys()):
            teststr = qelem.lower()
            # control param
            if strict is True and \
               teststr not in ALLOWED_PARAMS and teststr not in CONTROL_PARAMS:
                continue
            # control language
            if language is not None and teststr in CONTROL_PARAMS:
                found_lang = str(qdict[qelem][0])
                if (language == 'de' and found_lang not in TARGET_LANG_DE) or \
                   (language == 'en' and found_lang not in TARGET_LANG_EN) or \
                   found_lang != language:
                    logging.debug('bad lang: %s %s %s', language, qelem, found_lang)
                    raise ValueError
            # insert
            newqdict[qelem] = qdict[qelem]
        newstring = urlencode(newqdict, doseq=True)
        parsed_url = parsed_url._replace(query=newstring)
    return parsed_url


def normalize_url(parsed_url, strict=False, language=None):
    '''Takes a URL string or a parsed URL and returns a (basically) normalized URL string'''
    if not isinstance(parsed_url, ParseResult):
        parsed_url = urlparse(parsed_url)
    # port
    if parsed_url.port is not None and parsed_url.port in (80, 443):
        parsed_url = parsed_url._replace(netloc=NETLOC_RE.sub('', parsed_url.netloc))
    # path: https://github.com/saintamh/alcazar/blob/master/alcazar/utils/urls.py
    newpath = PATH1.sub('/', parsed_url.path)
    # Leading /../'s in the path are removed
    newpath = PATH2.sub('', newpath)
    # fragment
    if strict is True:
        newfragment = ''
    else:
        newfragment = parsed_url.fragment
    # lowercase + remove fragments
    parsed_url = parsed_url._replace(
                 scheme=parsed_url.scheme.lower(),
                 netloc=parsed_url.netloc.lower(),
                 path=newpath,
                 fragment=newfragment
                 )
    # strip unwanted query elements
    parsed_url = clean_query(parsed_url, strict, language)
    # rebuild
    return parsed_url.geturl()
